tailwind.config = {
    theme: {
      extend: {
        colors: {
          primary: {
            light: '#161b22',
            foreground: '#0a0f1a',
            DEFAULT: '#02050e'
          },
          accent: {
            DEFAULT: '#be9e6b'
          }
        }
      }
    }
  }